#!/usr/bin/perl -w
# senseLearner.pl
#
# Disambiguate all words in unrestricted text, based on information learned
# from sense tagged data (e.g. SemCor)
#
# Copyright (c) 2004-2005 Rada Mihalcea
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#
#
# INPUT:    file formatted using the SemCor XML format
# OUTPUT:   file annotated with word senses, also in SemCor XML format
#           [optionally evaluation figures]
#
# CHANGE LOG
#
# Date          Author          Change
# =======================================================================
# 01-20-2003    Rada            initial version
# 11-01-2004    Rada            add modules NNFeatures, VBFeatures, JJFeatures
#                               for additional flexibility
# 01-28-2005    Andras Csomai   add save,load, run parameters, to increase 
#                               speed and flexibility
# =======================================================================


use Getopt::Long;
use Socket;



# set default options and read command line options
if (defined $ENV{'SENSELEARNERHOME'}){
	$DEFAULT_DIR=$ENV{'SENSELEARNERHOME'};
	$DEFAULT_DIR=~s/\/$//;
	}
	else{
	$DEFAULT_DIR='~/SenseLearner';
	}
#'
require "$DEFAULT_DIR/Word.pm";
require "$DEFAULT_DIR/Models.pm";


$WINDOW_SIZE = 3;
$MODELS_DIR = "$DEFAULT_DIR/models";
$TIMBLE_DIR = "$DEFAULT_DIR/TimblModels";
if(scalar(@ARGV) == 0) {
    &displayHelp;
    exit;
}

&getCommandLineOptions(\$wordsBefore, \$wordsAfter, \$inFile, \$outFile, \$evaluate, \@models,\$Tsave, \$Tload, \$Trun, \$format, \$default, \$unload);
if (defined $unload){
     &killServers();
}


if (!defined($Tsave) && !defined($Tload) && !defined($Trun)){
   # The "classic" way, we do not create TIMBLE servers, but retrain TIMBLE at each run

   # check if training files exist
   &checkTrainingFiles(\@models);

   # read in text and store it in the @discourse strucure
   if ((defined $format) && ($format==1))
     {
        # read POS tagged file format
	&readinTextPlain(\@discourse, $inFile);
     }
    else
    {     # read SemCor-like format
 	&readinText(\@discourse, $inFile);
    }

  # apply SenseLearner algorithm
  &applySenseLearner(\@discourse, \@models);

  if(defined($evaluate)) {
  # evaluate the disambiguation
    &evaluate(\@discourse);
  }

  # store the disambiguation results in output file
   if ((defined $format) && ($format==1))
     {
       &printResultsPlain(\@discourse, $outFile);
      }
     else
     {
        &printResults(\@discourse, $outFile);
      }

  # remove all intermediate files
  `rm $DEFAULT_DIR/tmp.test.*`;

  exit;
}

# and now, the "advanced" options, loading, training, saving and questionning Timbl servers.

 if (defined($Tsave)){
# check if training files exist
&checkTrainingFiles(\@models);
}

if (defined($Tload)){
#checking if the required Timbl representation files exist
&checkTimbleFiles(\@models);
}

if (defined($Trun)){
# read in text and store it in the @discourse strucure
if ((defined $format) && ($format==1))
   {
        # read POS tagged file format
	&readinTextPlain(\@discourse, $inFile);
   }
 else
  {     # read SemCor-like format
	&readinText(\@discourse, $inFile);      
  }

}

# apply SenseLearner algorithm  -it vill act according to the save, load, run preferences
&applySenseLearnerPlain(\@discourse, \@models);

if (defined($Trun)){
if(defined($evaluate)) {
# evaluate the disambiguation
    &evaluate(\@discourse);}


# store the disambiguation results in output file
 if ((defined $format) && ($format==1))
    {
      &printResultsPlain(\@discourse, $outFile);
    }
   else
    {
      &printResults(\@discourse, $outFile);
    }

# remove all intermediate files
`rm $DEFAULT_DIR/tmp.test.*`;





########################################################################
#
#  S U B R O U T I N E S
#
########################################################################


# ======================================================================
# checkTrainingFiles
#
# checks on the existence of training files
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-01-2004    Rada            Initial version
#=======================================================================
sub checkTrainingFiles {
    my($modelsRef) = @_;

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);

    for(my $i=0; $i < $lenModels; $i++) {
	(-e "$MODELS_DIR/training.".$models[$i]) || die "Could not open training file $MODELS_DIR/training.".$models[$i];
    }
}
sub checkTimbleFiles {
    my($modelsRef) = @_;

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);

    for(my $i=0; $i < $lenModels; $i++) {
	(-e "$TIMBLE_DIR/tmp.savedtree.".$models[$i]) || die "Could not open Timbl model file $TIMBLE_DIR/tmp.savedtree.".$models[$i];
	(-e "$TIMBLE_DIR/tmp.savedtree.".$models[$i].".wgt") || die "Could not open Timbl model file $TIMBLE_DIR/tmp.savedtree.".$models[$i].".wgt";
	(-e "$TIMBLE_DIR/tmp.wordlist.".$models[$i]) || die "Could not open Timbl model file $TIMBLE_DIR/tmp.savedtree.".$models[$i];
    }
}
# ======================================================================
# applySenseLearner
#
# apply the sense learner algorithm
# - build feature vectors for all adj, nouns, verbs
# and do a learning using the given training files
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-22-2004    Rada            Initial version
# 11-01-2004    Rada            Add calls to extractFeatures functions
#=======================================================================
sub applySenseLearner {
    my ($discourseRef, $modelsRef) = @_;

    my @discourse = @$discourseRef;
    my $lenDiscourse = scalar(@discourse);

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);


    # EXTRACT FEATURES, ACCORDING TO EACH MODEL
    # and store them in tmp.test.xx files
    
    my @fileHandles;

    # open files to store feature vectors
    for(my $m=0; $m < $lenModels; $m++) {
	open $fileHandles[$m], ">$DEFAULT_DIR/tmp.test.$models[$m]";
	(-w $fileHandles[$m]) || die "Could not open test file $DEFAULT_DIR/tmp.test.$models[$m]\n";
    }


    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};

	# extract features for all models for this sentence
	for(my $m=0; $m < $lenModels; $m++) {
	    &{$models[$m]}(\@sentence, $fileHandles[$m], $wordsBefore, $wordsAfter);
	}

    }

    # close files
    for(my $m=0; $m < $lenModels; $m++) {
	close $fileHandles[$m];
    }



    # LEARNING STEP
    # Invoke the MBL learner through system calls
    for(my $m=0; $m < $lenModels; $m++) {
	`Timbl -f $MODELS_DIR/training.$models[$m] -t $DEFAULT_DIR/tmp.test.$models[$m] -o $DEFAULT_DIR/tmp.test.$models[$m].mbl`;
    }

    # And collect the predictions made by the learner into a single file
    # Make sure the concept predicted is the same as the target word
    my $awkcommand = "awk '{print \$1\" \"\$2\" \"\$NF}' ";
    for(my $m=0; $m < $lenModels; $m++) {
	$awkcommand .= "$DEFAULT_DIR/tmp.test.$models[$m].mbl ";
    }
    $awkcommand .= " | sed 's/[\,\#]/ /g' | awk '{if (\$2 == \$3) print \$1\" \"\$4}' | sort | uniq > $DEFAULT_DIR/tmp.test.results.mbl";
    `$awkcommand` ;


    # Read the predicted senses
    open PREDICTIONS, "<$DEFAULT_DIR/tmp.test.results.mbl";
    (-r PREDICTIONS) || die "Could not open result file with sense predictions\n";

    while($line = <PREDICTIONS>) {
	chomp $line;
	my ($id, $sense) = split /\s+/, $line;

	# make sure only *one* label is provided for an instance
	# if several labels are available, it means that various models
	# did not agree on the classification, and thus no prediction 
	# is made
	if(exists($predictions{$id})) {
	    delete($predictions{$id});
	}
	
	else {
	    $predictions{$id} = $sense;
	}
    }
    close PREDICTIONS;

    
    # APPLY THE PREDICTIONS ON THE TEST FILE
    
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};

	my $lenSentence = scalar(@sentence);

	for(my $i=0; $i < $lenSentence; $i++) {
	    if($sentence[$i]->{type} eq "wsd") {
		if(exists $predictions{$sentence[$i]->{id}}) {
		    $sentence[$i]->{predicted} = $predictions{$sentence[$i]->{id}};
		}
	    }
	}
	@{$discourse[$d]} = @sentence;
    }
    
    @$discourseRef = @discourse;
}

# ======================================================================
# applySenseLearnerPlain
#
# apply the sense learner algorithm for the enhanced mode, 
# - build feature vectors for all adj, nouns, verbs
# and do a learning using the given training files
#   train and load TIMBLE instances
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-22-2004    Rada            Initial version
# 11-01-2004    Rada            Add calls to extractFeatures functions
#=======================================================================
sub applySenseLearnerPlain {
    my ($discourseRef, $modelsRef) = @_;

    my @discourse = @$discourseRef;
    my $lenDiscourse = scalar(@discourse);

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);
if (defined $Trun){

    # EXTRACT FEATURES, ACCORDING TO EACH MODEL
    # and store them in tmp.test.xx files
    my @fileHandles;
    # open files to store feature vectors

    for(my $m=0; $m < $lenModels; $m++) {
	open $fileHandles[$m], ">$DEFAULT_DIR/tmp.test.$models[$m]";
	(-w $fileHandles[$m]) || die "Could not open test file $DEFAULT_DIR/tmp.test.$models[$m]\n";
    }

    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};
	# extract features for all models for this sentence
	for(my $m=0; $m < $lenModels; $m++) {
	    &{$models[$m]}(\@sentence, $fileHandles[$m], $wordsBefore, $wordsAfter);
	}

    }

    # close files
    for(my $m=0; $m < $lenModels; $m++) {
	close $fileHandles[$m];
    }

    }


    # LEARNING STEP
    # Invoke the MBL learner through system calls
    # depending on the command line argument, we either train the Timbl and save the output into the temporary files (-s),
	# either load the trained model (-l)
	# either run the loaded test with the trained model.
	my $modelList="";
	my $actPort;
	my $wordlistFiles="";
	if (defined $Tload)  {
	    for(my $m=0; $m < $lenModels; $m++) {
		$actPort=$Tload+$m;
		$modelList=$modelList." -model ".$models[$m];
		`Timbl -S $actPort +v s -i $TIMBLE_DIR/tmp.savedtree.$models[$m] >$DEFAULT_DIR/tmp.test.timble&`;
		$wordlistFiles=$wordlistFiles."$TIMBLE_DIR/tmp.wordlist.$models[$m]\n";
		}
	    print "\n----------------\n Initialization completed, wait for timble completion messages\n\n\n";
		open $modelFile,">$DEFAULT_DIR/startedModels.sln";
		print $modelFile $modelList,"\n",$lenModels,"\n",$wordlistFiles;
		close $modelFile;
	    exit;
	}
	elsif (defined $Tsave){
		print "evaluating testfile \n";
	    	for(my $m=0; $m < $lenModels; $m++) {
			`Timbl -f $MODELS_DIR/training.$models[$m] -I $TIMBLE_DIR/tmp.savedtree.$models[$m]`;
			my %knownWords;
			open(INWORD,"<$MODELS_DIR/training.$models[$m]") or die "could not open model file $MODELS_DIR/training.$models[$m]\n";
			while (<INWORD>){
			@parts=split(/ /,);
			$knownWords{$parts[1]}=1;
			}
			close INWORD;
			open(OUTWORD,">$TIMBLE_DIR/tmp.wordlist.$models[$m]") or die "could not open model file $TIMBLE_DIR/tmp.wordlist.$models[$m]\n";
			foreach (keys %knownWords){
			print OUTWORD "$_\n";
			}
			close OUTWORD;
 	   }
	   print "training over, results saved\n";
	  exit;
	}
   elsif (defined $Trun){

 	 my $strm;
	my $response;
      for(my $m=0; $m < $lenModels; $m++) {
#	`Timbl -f $MODELS_DIR/training.$models[$m] -t $DEFAULT_DIR/tmp.test.$models[$m] -o $DEFAULT_DIR/tmp.test.$models[$m].mbl`;
	use IO::Socket;
  	$actport=$Trun;
	my $sock ;
	$sock = IO::Socket::INET->new(
                        Proto    => "tcp",
                        PeerAddr => "127.0.0.1",
                        PeerPort => $actport,
                    )
                  or die "cannot connect to buffServ";

	 if (!(open $inputF , "<$DEFAULT_DIR/tmp.test.$models[$m]")){
		die "could not open input file $DEFAULT_DIR/tmp.test.$models[$m]\n";
	}
	 open $outpF , ">$DEFAULT_DIR/tmp.test.$models[$m].mbl";

	# read the greeting from buffServ
	 $strm=<$sock>;
	# now feed each line of the testfile into Timbl
	while ($line=<$inputF>){
		chomp $line;
		print $sock "c $m $line \n";
		#print "Client->Server:>c $m $line \n";
		#read buffserv's responses;
		$response="";
   	        while (( !($response=~/ERROR/ ) && (!($response=~/CATEGORY/ ))) || ($response eq "")) {
			$response=<$sock>;
			#print "Server->Client:>$response\n---\n";
			if ($response=~/CATEGORY/){
			$pred=$response;
			chomp($pred);
			$pred=~s/\s*CATEGORY \{(.*)\}\s*/$1/;
			print $outpF "$line $pred\n";
			}
	        }

	}
       print $sock "exit\n";
       close($sock);
	close $outpF;

    }
    # And collect the predictions made by the learner into a single file
    # Make sure the concept predicted is the same as the target word
    my $awkcommand = "awk '{print \$1\" \"\$2\" \"\$NF}' ";
    for(my $m=0; $m < $lenModels; $m++) {
	$awkcommand .= "$DEFAULT_DIR/tmp.test.$models[$m].mbl ";
    }
    $awkcommand .= " | sed 's/[\,\#]/ /g' | awk '{if (\$2 == \$3) print \$1\" \"\$4}' | sort | uniq > $DEFAULT_DIR/tmp.test.results.mbl";
    `$awkcommand` ;


    # Read the predicted senses
    open PREDICTIONS, "<$DEFAULT_DIR/tmp.test.results.mbl";
    (-r PREDICTIONS) || die "Could not open result file with sense predictions $DEFAULT_DIR/tmp.test.results.mbl\n";

    while($line = <PREDICTIONS>) {
	chomp $line;
	my ($id, $sense) = split /\s+/, $line;

	# make sure only *one* label is provided for an instance
	# if several labels are available, it means that various models
	# did not agree on the classification, and thus no prediction
	# is made
	if(exists($predictions{$id})) {
	    delete($predictions{$id});
	}

	else {
	    $predictions{$id} = $sense;
	}
    }
    close PREDICTIONS;


    # APPLY THE PREDICTIONS ON THE TEST FILE

    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};

	my $lenSentence = scalar(@sentence);

	for(my $i=0; $i < $lenSentence; $i++) {
	    if($sentence[$i]->{type} eq "wsd") {
		if(exists $predictions{$sentence[$i]->{id}}) {
		    $sentence[$i]->{predicted} = $predictions{$sentence[$i]->{id}};
		}
	    }
	}
	@{$discourse[$d]} = @sentence;
    }

    @$discourseRef = @discourse;

	}


}

# ======================================================================
# readinText
#
# read all sentences
# store Word instances for each word, punctuation
# the resulting structure (@discourse) is a matrix
#  - sentences are on rows
#  - for each sentence (row) words are on columns
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-03-2004    Rada            initial version
#=======================================================================

# a global variable incremented for each instance -- represents
# a unique id to be assigned to those instances that do not have
# an id in the input file
$wordId = 0;

sub readinText {
    my ($discourseRef, $inFile) = @_;
    my @discourse = @$discourseRef;

    my $sentIndex = 0;
    my $wordIndex = 0;


    # load input data
    # open the SemCor-like file, and start reading it
    open INFILE, "<$inFile" || die "Could not open input file $inFile. Please check path and make sure is valid\n";
    (-r INFILE) || die "Could not open input file $inFile. Please check path and make sure is valid\n";

    $testline=<INFILE>;
    if (!($testline=~/\<.*=.*\>/)){
    print "\n \n WARNING:  Most probably you have specifeid the wrong input format!!!! \n \n";
    }
    close INFILE;
    open INFILE, "<$inFile" || die "Could not open input file $inFile. Please check path and make sure is valid\n";
    (-r INFILE) || die "Could not open input file $inFile. Please check path and make sure is valid\n";

    while(my $line = <INFILE>) {
	chomp $line;
	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	# if end of sentence, reset wordIndex, and increase sentIndex
	if($line =~ /^\<\/s\>/) {
	    $wordIndex=0;
	    $sentIndex++;
	    next;
	}

	# if inside sentence, store Word-s
	if($line =~ /^\<wf/ ||
	   $line =~ /^\<punc/) {
	    $word = new Word($line);
	    if($word->{id} eq "?") {
		$wordId++;
		$word->{id} = "AUTO.".$wordId;
	    }
	    if(defined $word) {
		$discourse[$sentIndex][$wordIndex++] = $word;
	    }
	}
    }

    @$discourseRef = @discourse;
}

# ======================================================================
# readinTextPlain
#
# read all sentences
# store Word instances for each word, punctuation
# the resulting structure (@discourse) is a matrix
#  - sentences are on rows
#  - for each sentence (row) words are on columns
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 04-18-2005    Andras Csomai   initial version
#=======================================================================

$wordId = 0;

sub readinTextPlain {
    my ($discourseRef, $inFile) = @_;
    my @discourse = @$discourseRef;

    my $sentIndex = 0;
    my $wordIndex = 0;


    # load input data
    # open the POS tagged file, and start reading it
    open INFILE, "<$inFile";
    (-r INFILE) || die "Could not open input file $inFile. Please check path and make sure is valid\n";

    $testline=<INFILE>;
    if ($testline=~/\<.*=.*\>/){
    print "\n \n WARNING:  Most probably you have specifeid the wrong input format!!!! \n \n";
    }
    close INFILE;
    open INFILE, "<$inFile";
    (-r INFILE) || die "Could not open input file $inFile. Please check path and make sure is valid\n";

    while(my $line = <INFILE>) {
	chomp $line;
	if (!($line eq "")){
	$line =~s/\s+/ /;
	@words=split/\s/,$line;
	# if inside sentence, store Word-s
	for ($i=0;$i<scalar(@words);$i++){
	if($words[$i] =~ /\/.*\//) {
	    $word = newplain Word($words[$i]);
	    if($word->{id} eq "?") {
		$wordId++;
		$word->{id} = "AUTO.".$wordId;
	    }
	    if(defined $word) {
		$discourse[$sentIndex][$wordIndex++] = $word;
	    }
	}
	}

        $wordIndex=0;
        $sentIndex++;
	}
    }

    @$discourseRef = @discourse;
}

#=======================================================================
# printResults
#
# print the results of the disambiguation program, in a Semcor-like
# format
#
# Note! It also prints out the lexsn field -- but this slows
# down the program quite a bit, since it loads index.sense in memory
# Remove the section for adding lexsn if not needed
#
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-03-2004    Rada            initial version
#=======================================================================

sub printResults {
    my ($discourseRef, $outFile) = @_;
    my @discourse = @$discourseRef;

    #prepare the output file for writing
    open OUTFILE, ">$outFile";
    (-w OUTFILE) || die "Could not open output file for writing $outFile.\n Please check path and make sure is valid\n";

    my $lenDiscourse = scalar(@discourse);
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sent = @{$discourse[$d]};


	if($d != 0) {
	    print OUTFILE "</s>\n";
	}
	print OUTFILE "<s snum=", $d+1, ">\n";

	my $lenSentence = scalar(@sent);
	for(my $i=0; $i < $lenSentence; $i++) {

	    # open class word
	    if($sent[$i]->{type} eq "wsd") {
		print OUTFILE "<wf cmd=done ";
		if($sent[$i]->{id} !~ "AUTO") {
		    print OUTFILE "id=", $sent[$i]->{id}," ";
		}
		print OUTFILE "pos=",$sent[$i]->{pos}," ";
		print OUTFILE "lemma=",$sent[$i]->{lemma}," ";
		print OUTFILE "wnsn=",$sent[$i]->{wnsn}," ";
		print OUTFILE "predicted=",$sent[$i]->{predicted};
		    if ($sent[$i]->{predicted}==0 && $default){
			   $sent[$i]->{predicted}=1;
			}

		print OUTFILE ">",$sent[$i]->{word}, "</wf>\n";
	    }

	    elsif ($sent[$i]->{type} eq "punct") {
		print OUTFILE "<punc>",$sent[$i]->{word},"</punc>\n";
	    }
	    else {
		print OUTFILE "<wf cmd=ignore ";
		print OUTFILE "pos=",$sent[$i]->{pos}," ";
		print OUTFILE "lemma=",lc($sent[$i]->{word});
		print OUTFILE ">",$sent[$i]->{word}, "</wf>\n";
	    }
	}

    }


    print OUTFILE "</s>\n";
    close OUTFILE;

}
}
#=======================================================================
# printResultsPlain
#
# print the results of the disambiguation program, in the POS tagged format
# format
#
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 04-18-2005    Andras Csomai   initial version
#=======================================================================

sub printResultsPlain {
    my ($discourseRef, $outFile) = @_;
    my @discourse = @$discourseRef;
    #prepare the output file for writing
    open OUTFILE, ">$outFile";
    (-w OUTFILE) || die "Could not open output file for writing $outFile.\n Please check path and make sure is valid\n";
    my $lenDiscourse = scalar(@discourse);

    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sent = @{$discourse[$d]};


	my $lenSentence = scalar(@sent);
	for(my $i=0; $i < $lenSentence; $i++) {
	    # open class word
	    print OUTFILE $sent[$i]->{word},"/";
	    if($sent[$i]->{type} eq "wsd") {
		if (($sent[$i]->{predicted}!=0) || ($default==1)){
		    if ($sent[$i]->{predicted}==0){
			   $sent[$i]->{predicted}=1;
			}
			print OUTFILE $sent[$i]->{lemma},"/",$sent[$i]->{pos};
			print OUTFILE "#",$sent[$i]->{predicted},' ';}
		   else{
                       print OUTFILE $sent[$i]->{lemma},"/",$sent[$i]->{pos}," ";
			}
	    }
	    else {
		print OUTFILE $sent[$i]->{lemma},"/";
		print OUTFILE $sent[$i]->{pos}," ";
	    }
	}

	print OUTFILE "\n";
    }

    close OUTFILE;
}




#=======================================================================
# evaluate
#
# evaluates the predicted senses versus the correctly assigned
# senses
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 02-27-2004    Rada            initial version
#=======================================================================

sub evaluate {
    my ($discourseRef) = @_;

    my @discourse = @$discourseRef;

    my $totalWords = 0;
    my $correctWords = 0;

    my $totalIdWords = 0;
    my $correctIdWords = 0;

    my $totalWordsAddressed = 0;
    my $totalIdWordsAddressed = 0;
    my $correctWordsAddressed = 0;
    my $correctIdWordsAddressed = 0;


    my $lenDiscourse = scalar(@discourse);
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sent = @{$discourse[$d]};

	my $lenSentence = scalar(@sent);

	for(my $i=0; $i < $lenSentence; $i++) {
	    if($sent[$i]->{type} eq "wsd" &&
	       $sent[$i]->{wnsn} != 0) {  # do not count words that cannot be evaluated

		# total words
		$totalWords++;
		if($sent[$i]->{predicted} > 0) {
		    $totalWordsAddressed++;
		}
		if($sent[$i]->{id} !~ "AUTO") {
		    $totalIdWords++;
		    if($sent[$i]->{predicted} > 0) {
			$totalIdWordsAddressed++;
		    }
		}
		
		# correctly tagged words -- words 
		# with no sense predicted (0) are considered
		# as tagged with sense #1
		if(($sent[$i]->{predicted} > 0 &&
		    $sent[$i]->{wnsn} == $sent[$i]->{predicted}) ||
		   ($sent[$i]->{predicted} <= 0 &&
		    $sent[$i]->{wnsn} == 1)) {
		    $correctWords++;
		}
		if($sent[$i]->{predicted} > 0 &&
		   $sent[$i]->{wnsn} == $sent[$i]->{predicted}) {
		    $correctWordsAddressed++;
		}

		if($sent[$i]->{id} !~ "AUTO") {
		    if(($sent[$i]->{predicted} > 0 &&
			$sent[$i]->{wnsn} == $sent[$i]->{predicted}) ||
		       ($sent[$i]->{predicted} <= 0 &&
			$sent[$i]->{wnsn} == 1)) {
			$correctIdWords++;
		    }
		    if($sent[$i]->{predicted} > 0 &&
		       $sent[$i]->{wnsn} == $sent[$i]->{predicted}) {
			$correctIdWordsAddressed++;
		    }
		}
		
	    }
	}
    }
    
    print "Evaluation:\n";
    print "==========\n\n";
    print "ALL OPEN CLASS WORDS:\n\n";

    print "[partial coverage, only words addressed by model(s)]\n";
    if(0 != $totalWordsAddressed) {
	print "Total words addressed: $totalWordsAddressed\n";
	print "Correct words addressed: $correctWordsAddressed (",
	$correctWordsAddressed/$totalWordsAddressed,")\n\n";
    }
    
    print "[complete coverage, using a back-off to first sense]\n";    
    if(0 != $totalWords) {
	print "Total words: $totalWords\n";
	print "Correct words: $correctWords (",
	$correctWords/$totalWords,")\n";
    }

    print "\n\n=========\n\n";
    print "ALL WORDS MARKED WITH AN ID (words with an id=xxx field):\n\n";
    print "[partial coverage, only words addressed by model(s)]\n";
    if(0 != $totalIdWordsAddressed) {
	print "Total IDwords addressed: $totalIdWordsAddressed\n";
	print "Correct IDwords addressed: $correctIdWordsAddressed (",
	$correctIdWordsAddressed/$totalIdWordsAddressed,")\n\n";
    }

    print "[complete coverage, using a back-off to first sense]\n";
    if(0 != $totalIdWords) {
	print "Total IDwords: $totalIdWords\n";
	print "Correct IDwords: $correctIdWords (",
	$correctIdWords/$totalIdWords,")\n\n\n";
    }

}


#=======================================================================
# killServers
#
# Kills all running Timbl instances and the buffer server;
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 05-10-2005    Andras Csomai   introducing save, load, run
#=======================================================================

sub killServers {
   @proc=`ps -x -o '%p %a' --cols 1000 |grep 'Timbl .*-S'| grep -v grep| awk '{print \$1}'`;
   foreach (@proc){
      chomp($_);
      `kill -9 $_`;
   }
      @proc=`ps -x -o '%p %a' --cols 1000 |grep 'buffServ.pl .*-s'| grep -v grep|awk '{print \$1}'`;
   foreach (@proc){
      chomp($_);
      `kill -9 $_`;
   }
   print "\nKill signal sent to all Timbl servers, aborting execution.\n\nATTENTION: no other task was executed\n\n";
   exit;
}




#=======================================================================
# getCommandLineOptions
#
# Get the command lines options
# Check existence/validity of mandatory options
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-24-2003    Rada            initial version
# 01-28-2005    Andras Csomai   introducing save, load, run
#=======================================================================

sub getCommandLineOptions {
    my ($wordsBeforeRef,$wordsAfterRef, $sourceFileRef,
	$outputFileRef, $evaluateRef, $modelsRef, $save, $load, $run, $format, $default, $unload) = @_;

    # set default options
    $$wordsBeforeRef = $WINDOW_SIZE;
    $$wordsAfterRef = $WINDOW_SIZE;

$$default=0;
    # get the options
    GetOptions('help', \&displayHelp,
	       'B=i', $wordsBeforeRef,
	       'A=i', $wordsAfterRef,
               'i=s',$sourceFileRef,
               'o=s', $outputFileRef,
	       'eval', $evaluateRef,
	       'model=s', $modelsRef,
		'save', $save,
		'load=i',$load,
		'run=i', $run,
		'pos', $format,
		'default', $default,
		'kill', $unload
		);

     if (defined $$unload){
   	  killServers;
	}

    if ((!defined $$sourceFileRef)  &&((!defined $$save)&& (!defined $$load))){
	print STDERR "Please specify an input file, in SemCor format using the -i option\n";
	exit;
    }

    if((!defined $$outputFileRef) &&((!defined $$save) &&(!defined $$load))){
	print STDERR "Please specify an output file, using the -o option\n";
	exit;
    }
    if (((defined $$save)&&(defined $$load))||((defined $$save)&&(defined $$run))||((defined $$run)&&(defined $$load))){
	print STDERR "Only one action can be performed at a time (save/load/run).\n";
	exit;
    }
    if ((defined $$format) && ($$format==1) && (defined $$evaluateRef)){
	print STDERR "Evaluation can only be done on SemCor-like format files.\n";
         exit;
	}
   if ((!defined @$modelsRef) && (!defined $$run)){
	print STDERR "Please specify the models.\n";
         exit;
	}
  if (defined $$run){
 	if (defined @$modelsRef){
 	print STDERR "Warning: the list of models is ignored, the disambiguation is done only for the words covered by the models currently running (see startedModel.sln). If you want to use a different set of models, you have to restart the servers corresponding to the new set of models.\n";
    	}
    # first look upt the list of started models from the startedModels.sln file
    open $modelFile,"<$DEFAULT_DIR/startedModels.sln" or die "List of started models startedModels.sln not found, you should start up the Timbl threads for the required models and the buffer server.\n";
      $modelList=<$modelFile>;
      chomp($modelList);
      close $modelFile;
      # we are filling up the list of models from the started model file
      $line=$modelList;
      $line=~s/^\s+//;
      $line=~s/\s+/ /g;
      $line=~s/-//g;
      my @modelnam;
      @modelnam=split/\s+/,$line;
      $modelnum=-1;
      for ($i=1;$i<scalar(@modelnam);$i++){
      	if (!($modelnam[$i] eq "model")){
      		$modelnum++;
      		$$modelsRef[$modelnum]=$modelnam[$i];
		print "$modelnam[$i]\n";
      	  }
        }
       }

}




#=================================================================
# displayHelp
#
# prints out the help message, including usage
#
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 02-19-2003    Rada            Initial version
# ========================================================================
sub displayHelp()
{
    print STDERR "Usage: senseLearner [-B N] [-A N] -i <input_file> -o <output-file> [-eval] -model <model> [-model <model>] \n\n";
    print STDERR "Minimally-supervised all words WSD\n\n";
    print STDERR "  -B N      the size of the context to the left of the target word,\n";
    print STDERR "            considered for feature extraction (default = 3)\n";
    print STDERR "  -A N      the size of the context to the right of the target word,\n";
    print STDERR "             considered for feature extraction (default = 3)\n";
    print STDERR "  -i <file> the input file, in SemCor format\n";
    print STDERR "  -o <file> the output file, to store the results\n";
    print STDERR "  -eval     run the evaluation, and output precision and coverage figures; \n";
    print STDERR "            the use of this option implies that correct word senses are\n";
    print STDERR "            available in the input file\n";
    print STDERR "  -model <model> the name of the model to be used for sense annotation; \n";
    print STDERR "            assumes the availability of a training instance base \n";
    print STDERR "            training.model and a subroutine that implements this model\n";
    print STDERR "  -save     train the TIMBLE instances, and save internal representations\n";
    print STDERR "  -load <port> load the already saved TIMBLE internal representations, starting at port <port>\n";
    print STDERR "  -run <port> run the SenseLearner, with TIMBLE servers opened at ports starting <port>\n";
    print STDERR "  -pos      if specified, reads POS tagged plain text. One sentence per line.  Otherwise reads SemCor-like format\n";
    print STDERR "  -default  if specified, assigns \"most frequent sense\" for the open class words not addressed by the system. \n";
    print STDERR "  -kill     if present, kills all running Timbl instances and the buffer server.\n";

}



